@@ -20,6 +20,10 @@
 #include <stdint.h>
 #include <string.h>
 #include <stdlib.h>
+
+#include <map>
+
+#include "trx_vty.h"
 #include "Logger.h"
 #include "Threads.h"
 #include "LMSDevice.h"
@@ -39,17 +43,74 @@ extern "C" {
 using namespace std;
 
 #define MAX_ANTENNA_LIST_SIZE 10
-#define LMS_SAMPLE_RATE GSMRATE*32
 #define GSM_CARRIER_BW 270000.0 /* 270kHz */
 #define LMS_MIN_BW_SUPPORTED 2.5e6 /* 2.5mHz, minimum supported by LMS */
 #define LMS_CALIBRATE_BW_HZ OSMO_MAX(GSM_CARRIER_BW, LMS_MIN_BW_SUPPORTED)
 #define SAMPLE_BUF_SZ    (1 << 20) /* Size of Rx timestamp based Ring buffer, in bytes */
 
-LMSDevice::LMSDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface, size_t chans, double lo_offset,
+
+/* Device Name Prefixes as presented by LimeSuite API LMS_GetDeviceInfo(): */
+#define LMS_DEV_SDR_USB_PREFIX_NAME "LimeSDR-USB"
+#define LMS_DEV_SDR_MINI_PREFIX_NAME "LimeSDR-Mini"
+#define LMS_DEV_NET_MICRO_PREFIX_NAME "LimeNET-Micro"
+
+/* Device parameter descriptor */
+struct dev_desc {
+	/* Does LimeSuite allow switching the clock source for this device?
+	 * LimeSDR-Mini does not have switches but needs soldering to select
+	 * external/internal clock. Any call to LMS_SetClockFreq() will fail.
+	 */
+	bool clock_src_switchable;
+	/* Does LimeSuite allow using REF_INTERNAL for this device?
+	 * LimeNET-Micro does not like selecting internal clock
+	 */
+	bool clock_src_int_usable;
+	/* Device specific maximum tx levels selected by phasenoise measurements, in dB */
+	double max_tx_gain;
+	/* Sample rate coef (without having TX/RX samples per symbol into account) */
+	double rate;
+	/* Sample rate coef (without having TX/RX samples per symbol into account), if multi-arfcn is enabled */
+	double rate_multiarfcn;
+	/* Coefficient multiplied by TX sample rate in order to shift Tx time */
+	double ts_offset_coef;
+	/* Coefficient multiplied by TX sample rate in order to shift Tx time, if multi-arfcn is enabled */
+	double ts_offset_coef_multiarfcn;
+	/* Device Name Prefix as presented by LimeSuite API LMS_GetDeviceInfo() */
+	std::string name_prefix;
+};
+
+static const std::map<enum lms_dev_type, struct dev_desc> dev_param_map {
+	{ LMS_DEV_SDR_USB,   { true,  true,  73.0, GSMRATE, MCBTS_SPACING, 8.9e-5, 7.9e-5, LMS_DEV_SDR_USB_PREFIX_NAME } },
+	{ LMS_DEV_SDR_MINI,  { false, true,  66.0, GSMRATE, MCBTS_SPACING, 8.9e-5, 8.2e-5, LMS_DEV_SDR_MINI_PREFIX_NAME } },
+	{ LMS_DEV_NET_MICRO, { true,  false, 71.0, GSMRATE, MCBTS_SPACING, 8.9e-5, 7.9e-5, LMS_DEV_NET_MICRO_PREFIX_NAME } },
+	{ LMS_DEV_UNKNOWN,   { true,  true,  73.0, GSMRATE, MCBTS_SPACING, 8.9e-5, 7.9e-5, "UNKNOWN" } },
+};
+
+static enum lms_dev_type parse_dev_type(lms_device_t *m_lms_dev)
+{
+	std::map<enum lms_dev_type, struct dev_desc>::const_iterator it = dev_param_map.begin();
+
+	const lms_dev_info_t* device_info = LMS_GetDeviceInfo(m_lms_dev);
+
+	while (it != dev_param_map.end())
+	{
+		enum lms_dev_type dev_type = it->first;
+		struct dev_desc desc = it->second;
+
+		if (strncmp(device_info->deviceName, desc.name_prefix.c_str(), desc.name_prefix.length()) == 0) {
+			LOGC(DDEV, INFO) << "Device identified as " << desc.name_prefix;
+			return dev_type;
+		}
+		it++;
+	}
+	return LMS_DEV_UNKNOWN;
+}
+
+LMSDevice::LMSDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface, size_t chan_num, double lo_offset,
 		     const std::vector<std::string>& tx_paths,
 		     const std::vector<std::string>& rx_paths):
-	RadioDevice(tx_sps, rx_sps, iface, chans, lo_offset, tx_paths, rx_paths),
-	m_lms_dev(NULL)
+	RadioDevice(tx_sps, rx_sps, iface, chan_num, lo_offset, tx_paths, rx_paths),
+	m_lms_dev(NULL), started(false), m_dev_type(LMS_DEV_UNKNOWN)
 {
 	LOGC(DDEV, INFO) << "creating LMS device...";
 
@@ -59,6 +120,11 @@ LMSDevice::LMSDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface, size_t c
 	tx_gains.resize(chans);
 
 	rx_buffers.resize(chans);
+
+	/* Set up per-channel Rx timestamp based Ring buffers */
+	for (size_t i = 0; i < rx_buffers.size(); i++)
+		rx_buffers[i] = new smpl_buf(SAMPLE_BUF_SZ / sizeof(uint32_t));
+
 }
 
 LMSDevice::~LMSDevice()
@@ -93,7 +159,7 @@ static void lms_log_callback(int lvl, const char *msg)
 	if ((unsigned int) lvl >= ARRAY_SIZE(lvl_map))
 		lvl = ARRAY_SIZE(lvl_map)-1;
 
-	LOGLV(DLMS, lvl_map[lvl]) << msg;
+	LOGLV(DDEVDRV, lvl_map[lvl]) << msg;
 }
 
 static void print_range(const char* name, lms_range_t *range)
@@ -110,7 +176,7 @@ static void print_range(const char* name, lms_range_t *range)
 int info_list_find(lms_info_str_t* info_list, unsigned int count, const std::string &args)
 {
 	unsigned int i, j;
-	vector<string> filters;
+	std::vector<string> filters;
 
 	filters = comma_delimited_to_vector(args.c_str());
 
@@ -134,11 +200,11 @@ int info_list_find(lms_info_str_t* info_list, unsigned int count, const std::str
 int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 {
 	lms_info_str_t* info_list;
-	const lms_dev_info_t* device_info;
 	lms_range_t range_sr;
 	float_type sr_host, sr_rf;
 	unsigned int i, n;
 	int rc, dev_id;
+	struct dev_desc dev_desc;
 
 	LOGC(DDEV, INFO) << "Opening LMS device..";
 
@@ -175,19 +241,20 @@ int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 
 	delete [] info_list;
 
-	device_info = LMS_GetDeviceInfo(m_lms_dev);
+	m_dev_type = parse_dev_type(m_lms_dev);
+	dev_desc = dev_param_map.at(m_dev_type);
 
 	if ((ref != REF_EXTERNAL) && (ref != REF_INTERNAL)){
 		LOGC(DDEV, ERROR) << "Invalid reference type";
 		goto out_close;
 	}
 
-	/* if reference clock is external setup must happen _before_ calling LMS_Init */
-	/* FIXME make external reference frequency configurable */
+	/* if reference clock is external, setup must happen _before_ calling LMS_Init */
 	if (ref == REF_EXTERNAL) {
 		LOGC(DDEV, INFO) << "Setting External clock reference to 10MHz";
-		/* Assume an external 10 MHz reference clock */
-		if (LMS_SetClockFreq(m_lms_dev, LMS_CLOCK_EXTREF, 10000000.0) < 0)
+		/* FIXME: Assume an external 10 MHz reference clock. make
+		   external reference frequency configurable */
+		if (!do_clock_src_freq(REF_EXTERNAL, 10000000.0))
 			goto out_close;
 	}
 
@@ -197,22 +264,13 @@ int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 		goto out_close;
 	}
 
-	/* LimeSDR-Mini does not have switches but needs soldering to select external/internal clock */
-	/* LimeNET-Micro also does not like selecting internal clock*/
-	/* also set device specific maximum tx levels selected by phasenoise measurements*/
-	if (strncmp(device_info->deviceName,"LimeSDR-USB",11) == 0){
-		/* if reference clock is internal setup must happen _after_ calling LMS_Init */
-		/* according to lms using LMS_CLOCK_EXTREF with a frequency <= 0 is the correct way to set clock to internal reference*/
-		if (ref == REF_INTERNAL) {
-			LOGC(DDEV, INFO) << "Setting Internal clock reference";
-			if (LMS_SetClockFreq(m_lms_dev, LMS_CLOCK_EXTREF, -1) < 0)
-				goto out_close;
-		}
-		maxTxGainClamp = 73.0;
-	} else if (strncmp(device_info->deviceName,"LimeSDR-Mini",12) == 0)
-		maxTxGainClamp = 66.0;
-	else
-		maxTxGainClamp = 71.0; /* "LimeNET-Micro", etc FIXME pciE based LMS boards?*/
+	/* if reference clock is internal, setup must happen _after_ calling LMS_Init */
+	if (ref == REF_INTERNAL) {
+		LOGC(DDEV, INFO) << "Setting Internal clock reference";
+		/* Internal freq param is not used */
+		if (!do_clock_src_freq(REF_INTERNAL, 0))
+			goto out_close;
+	}
 
 	/* enable all used channels */
 	for (i=0; i<chans; i++) {
@@ -227,16 +285,22 @@ int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 		goto out_close;
 	print_range("Sample Rate", &range_sr);
 
-	LOGC(DDEV, INFO) << "Setting sample rate to " << GSMRATE*tx_sps << " " << tx_sps;
-	if (LMS_SetSampleRate(m_lms_dev, GSMRATE*tx_sps, 32) < 0)
+	if (iface == MULTI_ARFCN)
+		sr_host = dev_desc.rate_multiarfcn * tx_sps;
+	else
+		sr_host = dev_desc.rate * tx_sps;
+	LOGC(DDEV, INFO) << "Setting sample rate to " << sr_host << " " << tx_sps;
+	if (LMS_SetSampleRate(m_lms_dev, sr_host, 32) < 0)
 		goto out_close;
 
 	if (LMS_GetSampleRate(m_lms_dev, LMS_CH_RX, 0, &sr_host, &sr_rf))
 		goto out_close;
 	LOGC(DDEV, INFO) << "Sample Rate: Host=" << sr_host << " RF=" << sr_rf;
 
-	/* FIXME: make this device/model dependent, like UHDDevice:dev_param_map! */
-	ts_offset = static_cast<TIMESTAMP>(8.9e-5 * GSMRATE * tx_sps); /* time * sample_rate */
+	if (iface == MULTI_ARFCN)
+		ts_offset = static_cast<TIMESTAMP>(dev_desc.ts_offset_coef_multiarfcn * sr_host);
+	else
+		ts_offset = static_cast<TIMESTAMP>(dev_desc.ts_offset_coef * sr_host);
 
 	/* configure antennas */
 	if (!set_antennas()) {
@@ -244,13 +308,7 @@ int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 		goto out_close;
 	}
 
-	/* Set up per-channel Rx timestamp based Ring buffers */
-	for (size_t i = 0; i < rx_buffers.size(); i++)
-		rx_buffers[i] = new smpl_buf(SAMPLE_BUF_SZ / sizeof(uint32_t));
-
-	started = false;
-
-	return NORMAL;
+	return iface == MULTI_ARFCN ? MULTI_ARFCN : NORMAL;
 
 out_close:
 	LOGC(DDEV, FATAL) << "Error in LMS open, closing: " << LMS_GetLastErrorMessage();
@@ -344,6 +402,43 @@ bool LMSDevice::stop()
 	return true;
 }
 
+bool LMSDevice::do_clock_src_freq(enum ReferenceType ref, double freq)
+{
+	struct dev_desc dev_desc = dev_param_map.at(m_dev_type);
+	size_t lms_clk_id;
+
+	switch (ref) {
+	case REF_EXTERNAL:
+		lms_clk_id = LMS_CLOCK_EXTREF;
+		break;
+	case REF_INTERNAL:
+		if (!dev_desc.clock_src_int_usable) {
+			LOGC(DDEV, ERROR) << "Device type " << dev_desc.name_prefix
+					  << " doesn't support internal reference clock";
+			return false;
+		}
+		/* According to lms using LMS_CLOCK_EXTREF with a
+		   frequency <= 0 is the correct way to set clock to
+		   internal reference */
+		lms_clk_id = LMS_CLOCK_EXTREF;
+		freq = -1;
+		break;
+	default:
+		LOGC(DDEV, ERROR) << "Invalid reference type " << get_value_string(clock_ref_names, ref);
+		return false;
+	}
+
+	if (dev_desc.clock_src_switchable) {
+		if (LMS_SetClockFreq(m_lms_dev, lms_clk_id, freq) < 0)
+			return false;
+	} else {
+		LOGC(DDEV, INFO) << "Device type " << dev_desc.name_prefix
+				 << " doesn't support switching clock source through SW";
+	}
+
+	return true;
+}
+
 /* do rx/tx calibration - depends on gain, freq and bw */
 bool LMSDevice::do_calib(size_t chan)
 {
@@ -385,7 +480,7 @@ bool LMSDevice::do_filters(size_t chan)
 
 double LMSDevice::maxTxGain()
 {
-	return maxTxGainClamp;
+	return dev_param_map.at(m_dev_type).max_tx_gain;
 }
 
 double LMSDevice::minTxGain()
@@ -711,8 +806,9 @@ int LMSDevice::readSamples(std::vector < short *>&bufs, int len, bool * overrun,
 	for (size_t i = 0; i < rx_buffers.size(); i++) {
 		rc = rx_buffers[i]->read(bufs[i], len, timestamp);
 		if ((rc < 0) || (rc != len)) {
-			LOGC(DDEV, ERROR) << rx_buffers[i]->str_code(rc);
-			LOGC(DDEV, ERROR) << rx_buffers[i]->str_status(timestamp);
+			LOGCHAN(i, DDEV, ERROR) << rx_buffers[i]->str_code(rc) << ". "
+						<< rx_buffers[i]->str_status(timestamp)
+						<< ", (len=" << len << ")";
 			return 0;
 		}
 	}
